import configargparse
from pathlib import Path
import os
import sys

import fitanimate.plot as fap
import fitanimate.data as fad

import matplotlib.pyplot as plt
plt.rcdefaults()
import matplotlib.animation as animation
import matplotlib.gridspec as gridspec
import matplotlib as mpl
from cycler import cycler

import cartopy.crs as crs

def getFontSize( x, dpi ):
    # For 64 point font for 4k (x=3840,y=2160) @ 100 dpi
    return int(64* x/3840 * 100.0/dpi)

def main():
    videoFormats = {
        '240p': (426,240),
        '360p': (640,360),
        '480p': (720,480),
        '720p': (1280,720),
        '1080p': (1920,1080),
        '1440p': (2560,1440),
        '4k' : (3840,2160)
    }

    defaultFields = ['timestamp', 'temperature', 'heart_rate', 'lap', 'gears', 'altitude', 'grad', 'distance']
    defaultPlots = ['cadence', 'speed', 'power']

    parser = configargparse.ArgumentParser(default_config_files=
                                           [ os.path.join( str(Path.home()), '.config', 'fitanimate', '*.conf'),
                                             os.path.join( str(Path.home()), '.fitanimate.conf') ],
                                           formatter_class=configargparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        'infile', metavar='FITFILE', type=configargparse.FileType(mode='rb'),
        help='Input .FIT file (Use - for stdin).',
    )
    parser.add_argument(
        '--offset', type=float, default=0.0, help='Time offset (hours).'
    )
    parser.add_argument(
        '--show',    '-s', action='store_true', default=False, help='Show the animation on screen.'
    )
    parser.add_argument(
        '--num',    '-n', type=int, default=0, help='Only animate the first NUM frames.'
    )
    parser.add_argument(
        '--fields', type=str, action='append', default=defaultFields, help='Fit file variables to display as text.', choices=fap.RideText.supportedFields
    )
    parser.add_argument(
        '--plots', type=str, action='append', default=defaultPlots, help='Fit file variables to display as bar plot.', choices=fap.supportedPlots
    )
    parser.add_argument(
        '--no-elevation', action='store_true', default=False, help='Disable elevation plot.'
    )
    parser.add_argument(
        '--no-map', action='store_true', default=False, help='Disable map.'
    )
    parser.add_argument(
        '--outfile', '-o', type=str, default=None, help='Output filename.'
    )
    parser.add_argument(
        '--format', '-f', type=str, default='1080p', choices=videoFormats.keys(),
        help='Output video file resolution.'
    )
    parser.add_argument(
        '--dpi', '-d', type=int, default=100,
        help='Dots Per Inch. Probably shouldn\'t change.'
    )
    parser.add_argument(
        '--text-color', '-c', type=str, default='black',
        help='Text Color.'
    )
    parser.add_argument(
        '--plot-color', type=str, default='tab:blue',
        help='Plot Color.'
    )
    parser.add_argument(
        '--highlight-color', type=str, default='tab:red',
        help='Plot Highlight Color.'
    )
    parser.add_argument(
        '--alpha', type=float, default=0.3, help='Opacity of plots.'
    )
    parser.add_argument(
        '--vertical', '-v', action='store_true', default=False, help='Plot bars Verticaly.'
    )
    parser.add_argument(
        '--elevation-factor', '-e', type=float, default=5.0, help='Scale the elevation by this factor in the plot.'
    )
    parser.add_argument(
        '--test', '-t', action='store_true', help='Options for quick tests. Equivalent to "-s -f 360p".'
    )
    args = parser.parse_args()

    if args.test:
        args.format = '360p'
        args.show = True

    if len(args.plots) != len(defaultPlots): # The user specified plots, remove the defaults
        args.plots = args.plots[len(defaultPlots):]

    if len(args.fields) != len(defaultFields): # As above
        args.fields = args.fields[len(defaultFields):]

    fap.PlotBase.alpha = args.alpha
    fap.PlotBase.hlcolor = args.highlight_color

    x, y = videoFormats[args.format]

    plt.rcParams.update({
        'font.size': getFontSize(x,args.dpi),
        'figure.dpi': args.dpi,
        'text.color': args.text_color,
        'axes.labelcolor': args.text_color,
        'xtick.color': args.text_color,
        'ytick.color': args.text_color,
        'axes.prop_cycle': cycler('color', [args.plot_color])
    })

    fig = plt.figure(figsize=(x/args.dpi,y/args.dpi))

    # Elevation
    if args.no_elevation: # Don't make the elevation plot and remove related text
        for f in ['altitude','grad']:
            if f in  args.fields:
                args.fields.remove(f)

    else:
        gs_e  = gridspec.GridSpec(1,1)
        gs_e.update( left=0.6, right=1.0, top=1.0, bottom=0.8)
        a_e   = plt.subplot( gs_e[0,0] )

    # Map
    if args.no_map:
        f = 'distance'
        if f in args.fields:
            args.fields.remove(f)

    else:
        projection = crs.PlateCarree()
        gs_m  = gridspec.GridSpec(1,1)
        gs_m.update( left=0.6, right=1.0, top=0.8, bottom=0.4)
        a_m   = plt.subplot( gs_m[0,0], projection=projection  )

    # Bar
    gs_b  = gridspec.GridSpec(1,1)
    # If horizontal, size depends on the number of bars
    if args.vertical:
        height = 0.15
    else:
        height = 0.05*len(args.plots)

    gs_b.update( left=0.11, right=1.0, top=height, bottom=0.0)
    a_bar = plt.subplot( gs_b[0,0] )

    fig.patch.set_alpha(0.) # Transparant background

    # See https://adrian.pw/blog/matplotlib-transparent-animation/

    # Bar plots
    plotVars = []
    for p in args.plots:
        plotVars.append( fap.newPlotVar(p) )

    if args.vertical:
        gs_b.update( left=0.0, bottom=0.05, top=0.25)
        plotBar = fap.BarPlot( plotVars, a_bar )
    else:
        plotBar = fap.HBarPlot( plotVars, a_bar )

    plots = [plotBar]

    # Text data
    plots.append( fap.RideText( fig, args.fields ) )

    if not args.no_map:
        mp = fap.MapPlot(a_m, projection )
        plots.append(mp)

    if not args.no_elevation:
        ep = fap.ElevationPlot( a_e, args.elevation_factor )
        plots.append(ep)

    record_names = []
    for plot in plots:
        record_names += plot.ffNames

    # Remove duplicates
    record_names = list(dict.fromkeys(record_names))
    dataGen = fad.DataGen( fad.prePocessData(args.infile, record_names , int(args.offset*3600.0) ) )

    if not args.no_map:
        mp.DrawBasePlot( dataGen.lonArr, dataGen.latArr )

    if not args.no_elevation:
        ep.DrawBasePlot( dataGen.dArr, dataGen.aArr )

    # Check the dimensions of the map plot and move it to the edge/top
    if not args.no_map:
        dyOverDx = mp.getHeightOverWidth()
        gs_points = gs_m[0].get_position(fig).get_points()
        xmin = gs_points[0][0]
        ymin = gs_points[0][1]
        xmax = gs_points[1][0]
        ymax = gs_points[1][1]
        dx=xmax-xmin
        dy=ymax-ymin
        if dyOverDx>1.0: # Tall plot. Maintain gridspec height, change width
            dx_new = dx/dyOverDx
            xmin_new = xmax - dx_new
            gs_m.update(left=xmin_new)
        else: # Wide plot. Move up
            dy_new = dy * max(dyOverDx,0.6) # Don't scale to less that 60%... messes up for some reason
            ymin_new = ymax - dy_new
            gs_m.update(bottom=ymin_new)

    nData = dataGen.dataSet.nFrames()
    if args.num:
        nData = args.num

    # Time interval between frames in msec.
    inter = 1000.0/float(dataGen.dataSet.fps)
    anim=animation.FuncAnimation(fig, fad.run, dataGen, fargs=(fig,tuple(plots),), repeat=False,blit=False,interval=inter,save_count=nData)

    outf = os.path.splitext(os.path.basename(args.infile.name))[0] + '_overlay.mp4'
    if args.outfile:
        outf = args.outfile

    if not args.show:
        anim.save(outf, codec="png", fps=dataGen.dataSet.fps,
                  savefig_kwargs={'transparent': True, 'facecolor': 'none'})

    if args.show:
        plt.show()

if __name__ == '__main__':
    main()



import fitparse

def safeData(d,name=None):
    if d is None:
        return None

    if name and name in ['position_lat','position_long']:
        # Divide by 2^32/360.
        return d/11930464.7

    try:
        fd = float(d)
    except TypeError:
        return None

    return fd


class DataSet:
    # Only iterpolated these fast changing variables
    do_interpolate =  ['power','speed','cadence']
    def __init__(self):
        self.data = []
        self.intData = []
        self.fps = 10

    def addData(self, data ):
        if len(self.data) < 1:
            self.data.append( data )
            return True

        t_prev = int(self.data[-1]['timestamp'])
        dt = int(data['timestamp'])-t_prev
        if dt == 0:
            return True

        if dt<0:
            print('Negative time delta! Not adding data')
            return False

        self.data.append( data )
        return True

    def interpolateData(self):
        for i in range(len(self.data)-1):
            d0 = self.data[i]
            d1 = self.data[i+1]
            self.intData.append(d0)
            for j in range(1,self.fps):
                self._step = j
                dnew = {}
                for f in d0.keys():
                    if f in d1 and f in self.do_interpolate:
                        dnew[f] = self._interpolate(d0[f],d1[f],j)
                        dnew['interpolated'] = True

                self.intData.append( dnew )

    def nFrames(self):
        return self.fps * len(self.data)

    def _interpolate(self, v0, v1, step ):
        return ( (self.fps-step)*v0 + step*v1)/float(self.fps)

    def dump(self):
        for d in self.data:
            print(d)


def prePocessData( infile, record_names, timeoffset=None ):
    dataset = DataSet()
    ff = fitparse.FitFile( infile )

    for message in ff.get_messages(['record','lap','event']):
        data = {}
        message_name = message.as_dict()['name']
        if message_name == 'record':
            data['timestamp'] = int(message.get_value('timestamp').timestamp())
            if timeoffset:
                data['timestamp'] += timeoffset

            for f in record_names:
                d = safeData( message.get_value(f), f )
                if not (d is None):
                    data[f] = d

            ok = dataset.addData(data)
            if not ok:
                print( 'Problem adding data point. Not adding any more data.')
                dataset.interpolateData()
                return dataset

        elif message_name == 'lap' and len(dataset.data)>0:
            # Just append to the previous data
            dataset.data[-1]['lap'] = True

        elif ( message_name == 'event' and
               message.get_raw_value('gear_change_data') and
               len(dataset.data)>0 ):
            gears = '{}-{}'.format(message.get_value('front_gear'),
                                   message.get_value('rear_gear') )
            dataset.data[-1]['gears'] = gears

    dataset.interpolateData()
    return dataset

def run(data,fig,plots):
    for plot in plots:
        plot.update(data)

# Yeilds to first argument of run()
class DataGen():
    def __init__(self, dataSet ):
        self.dataSet = dataSet

        self.aArr = []
        self.dArr = []

        self.latArr = []
        self.lonArr = []

        for data in dataSet.data:
            if 'altitude' in data and 'distance' in data:
                self.aArr.append(data['altitude'])
                self.dArr.append(data['distance'])

            if 'position_lat' in data and 'position_long' in data:
                self.latArr.append(data['position_lat'])
                self.lonArr.append(data['position_long'])

        if len(self.aArr)>0:
            self.mkGradData()

    def mkGradData(self):
        # Smooth second-by-seceond altitude and distance data to get
        # better gradient estimates
        #
        # Easier to do this here instead of in preProcessData()
        # since we now have the altitude and distance arrrays

        a = []
        d = []
        window_size = 5
        i = 0
        if len(self.aArr) != len(self.dArr):
            print( 'Warning missmatch in distance and altitude data.')
            return

        while i < len(self.aArr) - window_size + 1:
            a.append(sum(self.aArr[i : i + window_size]) / window_size)
            d.append(sum(self.dArr[i : i + window_size]) / window_size)
            i+=1

        alast=None
        dlast=None
        glast=0.0
        grad=0.0
        g = []
        for i in range(len(a)):
            if not (dlast is None) and not(alast is None):
                dd = d[i]-dlast
                da = a[i]-alast
                if dd != 0.0:
                    grad = 100.0*da/dd
                else:
                    grad = glast
                g.append(grad)
            dlast = d[i]
            alast = a[i]
            glast = grad

        # Will be window_size-1 fewer entries. Pad the start.
        for i in range(window_size):
            g.insert(0,g[0])

        # Now insert the gradient data
        i = 0
        for data in self.dataSet.data:
            if 'altitude' in data and 'distance' in data:
                if i >= len(g):
                    print('Warning grad array size data missmatch.')
                    break

                data['grad'] = g[i]
                i+=1

    def __call__(self):
        for data in self.dataSet.intData:
            yield data



from datetime import datetime
import matplotlib.pyplot as plt


class TextLine:
    def __init__(self, fig, field_name, txt_format, x=None, y=None, scale=None ):
        self.fig = fig
        self.field_name = field_name
        self.txt_format = txt_format
        self.x = x
        self.y = y
        self.value = 0
        self.scale = scale

        self.fig_txt = None

    def setAxesText(self):
        if not self.fig_txt:
            self.fig_txt = self.fig.text( self.x, self.y, self.txt_format.format( self.value ) )
            return

        self.fig_txt.set_text( self.txt_format.format( self.value ) )

    def setValue(self, data ):
        # Don't update the text data if it is just a subsecond interpolation
        if 'interpolated' in data and data['interpolated']:
            return False

        if not (self.field_name in data):
            return False

        self.value = data[self.field_name]
        if self.scale:
            self.value *= self.scale

        return True

class CounterTextLine(TextLine):
    def __init__(self, fig, field_name, txt_format, x=None, y=None ):
        TextLine.__init__(self, fig, field_name, txt_format, x, y )

    def setValue(self, data):
        if self.value == 0 or self.field_name in data:
            self.value += 1
            return True

        return False

class TSTextLine(TextLine):
    def __init__(self, fig, field_name, txt_format, x=None, y=None,
                 timeformat='%H:%M:%S' ):
        TextLine.__init__(self, fig, field_name, txt_format, x, y )
        self.timeformat = timeformat

    def setValue(self,data):
        if not TextLine.setValue(self, data):
            return False

        self.value = datetime.fromtimestamp(int(self.value)).strftime(
            self.timeformat)
        return True


class TextPlot:
    def __init__(self, fig ):
        self.fig = fig
        self.textLines = []
        self._ffNames = []

        # Postion of first text object if not specified
        self.x = 0.02
        self.y = 0.95

        # If position for new text is not given offset from previous text by this much
        self.dx = 0.0
        self.dy = -0.06

    def addTextLine(self, textLine ):
        nlines = len(self.textLines)

        if nlines < 1:
            xprev = self.x-self.dx
            yprev = self.y-self.dy
        else:
            xprev = self.textLines[-1].x
            yprev = self.textLines[-1].y

        if textLine.x is None:
            textLine.x = xprev + self.dx

        if textLine.y is None:
            textLine.y = yprev + self.dy

        self.textLines.append( textLine )

        self._ffNames.append( textLine.field_name )

    @property
    def ffNames(self):
        """
        Return list of fit file record variable names requred for this plot
        """
        return self._ffNames

    def update(self, data):
        for txtLine in self.textLines:

            if not txtLine.setValue( data ):
                continue

            txtLine.setAxesText()

class RideText(TextPlot):
    supportedFields = ['timestamp', 'temperature', 'core_temperature', 'heart_rate', 'lap', 'gears', 'altitude', 'grad', 'distance']
    def __init__(self, fig, fields ):
        TextPlot.__init__(self, fig )
        self.fields = fields

        if 'timestamp' in self.fields:
            self.addTextLine( TSTextLine( self.fig,'timestamp', '{}' )) #, x=.1, y=.9 ))

        if 'temperature' in self.fields:
            self.addTextLine( TextLine( self.fig,'temperature', '{:.0f} ℃'))

        if 'core_temperature' in self.fields:
            self.addTextLine( TextLine( self.fig,'core_temperature', '{:.1f} ℃'))

        if 'heart_rate' in self.fields:
            self.addTextLine( TextLine( self.fig,'heart_rate',  '{:.0f} BPM'))

        if 'lap' in self.fields:
            self.addTextLine( CounterTextLine( self.fig, 'lap', 'Lap {}'))

        if 'gears' in self.fields:
            self.addTextLine( TextLine( self.fig, 'gears', '{}'))

        # Position near the elevation profile
        if 'altitude' in self.fields or 'grad' in self.fields:
            self.addTextLine( TextLine( self.fig, 'altitude','{:.0f} m', x=0.9, y=0.95) )
            self.addTextLine( TextLine( self.fig, 'grad', '{:5.1f}%'))

        # Near the map
        if 'distance' in self.fields:
            self.addTextLine( TextLine( self.fig, 'distance', '{:.1f} km', y=0.75, scale=0.001))

    # @property
    # def ffNames(self):
    #     """
    #     Return list of fit file record variable names requred for this plot
    #     """
    #     return [ 'temperature', 'altitude', 'heart_rate', 'gradient', 'distance']


# Information about a fitfile record to plot
class PlotVar:
    def __init__(self, ffname, name, units, maxValue, minValue=0.0, scaleFactor=1.0, offSet=0.0):
        self.ffname = ffname # name in fit file
        self.name = name
        self.units = units
        self.maxValue = maxValue
        self.minValue = minValue
        self.scaleFactor = scaleFactor # Multiply ff data by this
        self.offSet = offSet # Add this to ff data


    def getNameLabel( self ):
        return '{} ({})'.format(self.name,self.units)

    def getNormValue( self, data ):
        """ Between 0 and 1"""
        return (self.getValue(data) - self.offSet)/(self.maxValue-self.minValue)

    def getValue(self, data ):
        val = data[self.ffname]
        return val*self.scaleFactor + self.offSet

    def getValueUnits(self, value ):
        return '{:.0f} {:}'.format(value,self.units)

supportedPlots = ['cadence', 'speed', 'power', 'heart_rate', 'None']
def newPlotVar(variable):
    if variable == 'cadence':
        return PlotVar( variable,'Cadence', 'RPM', 120.0 )

    if variable == 'speed':
        return PlotVar('speed', 'Speed', 'km/h', 80.0, scaleFactor=3.6 )

    if variable == 'power':
        return PlotVar('power', 'Power',' W', 1000.0)

    if variable == 'heart_rate':
        return PlotVar('heart_rate', 'HeartRate',' BPM', 200.0)

    if variable == 'None':
        return None

    raise ValueError( 'Illegal variable {}. Must be one of: '.format(variable) + ' '.join([str(v) for v in self.supportedPlots]))

class PlotBase:
    alpha = 0.3
    hlcolor = 'tab:green'

    # Nominal marker sizes are for 3840x2160 (4K) at 100 DPI
    nom_dpi = 100.0
    nom_size = [3840/nom_dpi, 2160/nom_dpi]

    # Normal plot marker size. Diameter in pixels.
    nom_pms = 12

    def __init__(self):
        f = plt.gcf()
        dpi = f.dpi
        size = f.get_size_inches()

        # Scale by size and DPI
        self.pms = self.nom_pms * size[0]/self.nom_size[0] * dpi/self.nom_dpi

        # area is pi*r^2
        self.sms = 3.14159*(0.5*self.pms)**2

class BarPlotBase(PlotBase):
    def __init__(self, plotVars, axes, value = 0.0):
        PlotBase.__init__(self)
        self.plotVars = plotVars
        self.axes = axes
        self.axes.autoscale_view('tight')
        self.axes.set_axis_on()
        self.axes.tick_params(axis=u'both', which=u'both',length=0)
        for s in ['top','bottom','left','right']:
            self.axes.spines[s].set_visible(False)

        self.makeBar( [ pv.name for pv in self.plotVars ] )

        self.text = []
        for i in range(len(self.plotVars)):
            pv = self.plotVars[i]
            self.appendText(i)

    @property
    def ffNames(self):
        """
        Return list of fit file variable names requred for this plot
        """
        return [ pv.ffname for pv in self.plotVars ]
    def update(self, data ):
        for i in range(len(self.plotVars)) :
            pv = self.plotVars[i]
            if not (pv.ffname in data):
                continue

            value = pv.getValue(data)
            self.text[i].set_text( pv.getValueUnits(value) )

            # scale the value for the bar chart
            value = pv.getNormValue(data)
            self.setBarValue( self.bar[i], value )

    def setBarValue(self, bar, value ):
        pass

    def appendText(self, i ):
        pass

    def makeBars(self, names):
        pass

class BarPlot(BarPlotBase):
    txt_dx = -0.12
    txt_dy = 0.05
    def __init__(self, plotVars, axes, value = 0.0):
        BarPlotBase.__init__(self, plotVars, axes, value )
        self.axes.set_ylim( 0.0, 1.0 )
        self.axes.get_yaxis().set_visible(False)

    def makeBar(self, names ):
        self.bar = self.axes.bar( x = names, height = [0.0]*len(names), alpha=self.alpha )

    def setBarValue(self, bar, value ):
        bar.set_height( value )

    def appendText(self, i ):
        pv = self.plotVars[i]
        self.text.append( self.axes.text( i+self.txt_dx, self.txt_dy, pv.getValueUnits(0.0) ) )

class HBarPlot(BarPlotBase):
    txt_dx = 0.01
    txt_dy = -0.28
    def __init__(self, plotVars, axes, value = 0.0 ):
        BarPlotBase.__init__(self, plotVars, axes, value )
        self.axes.set_xlim( 0.0, 1.0 )
        self.axes.get_xaxis().set_visible(False)

    def makeBar(self, names ):
        self.bar = self.axes.barh( y = names, width = [0.0]*len(names), alpha=self.alpha )

    def setBarValue(self, bar, value ):
        bar.set_width( value )

    def appendText(self, i ):
        pv = self.plotVars[i]
        self.text.append( self.axes.text( self.txt_dx, i+self.txt_dy, pv.getValueUnits(0.0) ) )

class ElevationPlot(PlotBase):
    # vscale: Scale the elevation up by this much relative to the distance
    def __init__(self, axes, vScale = 5.0 ):
        PlotBase.__init__(self)
        self.axes = axes
        self.vScale = vScale

        self.axes.set_axis_off()
        for s in ['top','bottom','left','right']:
            self.axes.spines[s].set_visible(False)

        self.axes.set_aspect(self.vScale)
        self.axes.tick_params(axis=u'both', which=u'both',length=0)

    def DrawBasePlot( self, distArr, elevArr ):
        self.axes.plot(distArr,elevArr,marker='.',markersize=self.pms,alpha=self.alpha)

    def update(self,data):
        if 'distance' in data and 'altitude' in data:
            self.axes.plot(data['distance'],data['altitude'],color=self.hlcolor,marker='.',markersize=self.pms)

    @property
    def ffNames(self):
        return [ 'distance', 'altitude' ]

class MapPlot(PlotBase):
    def __init__(self, axes, projection ):
        PlotBase.__init__(self)
        self.axes = axes
        self.axes.outline_patch.set_visible(False)
        self.axes.background_patch.set_visible(False)
        self.projection = projection

    def DrawBasePlot( self, lonArr, latArr ):
        lon_min=min(lonArr)
        lon_max=max(lonArr)
        lat_min=min(latArr)
        lat_max=max(latArr)
        dlon = lon_max-lon_min
        dlat = lat_max-lat_min
        b=[ lon_min-0.02*dlon,
            lon_max+0.05*dlon,
            lat_min-0.02*dlat,
            lat_max+0.02*dlat ]
        self.axes.set_extent( b, crs=self.projection )
        self.axes.scatter( lonArr, latArr,s=self.sms,marker='.',alpha=self.alpha,transform=self.projection )

    def getHeightOverWidth(self):
        ymin,ymax = self.axes.get_ylim()
        dy=ymax-ymin
        xmin,xmax = self.axes.get_xlim()
        dx=xmax-xmin
        return dy/dx

    def update(self,data):
        if 'position_lat' in data and 'position_long' in data:
            self.axes.scatter(data['position_long'],data['position_lat'],color=self.hlcolor,marker='.',s=self.sms,alpha=self.alpha,transform=self.projection )

    @property
    def ffNames(self):
        return [ 'position_lat', 'position_long' ]
